/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("jazz.app.internal.AuthenticationInterceptor");

dojo.require("jazz.app.events");
dojo.require("jazz.app.internal.loginUtil");

// enumeration of possible authentication states
var STATE_NEW_CHALLENGE = 0;
var STATE_NORMAL_RESPONSE = 1;
var STATE_PREVIOUS_ATTEMPT_FAILED = 2;
var STATE_LOGGED_OUT = 3;

// variables whose states span requests
var challengePending = false;
var credentialProvider = null;

/**
 * Inspects the response to determine if an authentication challenge is required.
 * If it is, it will display a login prompt and submit the credentials submitted
 * by the user. Upon successful login, it will tell the request manager to send
 * any pending requests, including the original request that resulted in the
 * challenge.
 * 
 * @param {Object} origRequest the request that resulted in the current response
 * @param {Object} response the response to inspect
 * @param {Object} requestManager the request manager that sent the request
 */
jazz.app.internal.AuthenticationInterceptor = function(origRequest, response, requestManager) {
	
	if(!credentialProvider) {
		console.warn("No credential provider set; cannot perform authentication");
		return false;
	}
	
	var state = getAuthState(response);
	
	// First handle the responses that can simply pass through 
	if(state === STATE_NORMAL_RESPONSE) {
		return false;
	} else if(state === STATE_LOGGED_OUT) {
		dojo.publish(jazz.app.events.LOGOUT, [null]);
		return false;
	}
	
	// If we're still running, then this response contains an authentication
	// challenge to either:
	// 
	// - A normal request
	// - A failed authentication request
	//
	// TODO: Consider adding a request interceptor to catch outgoing requests
	// when we're in a pending state.
	
	// Use the default GUI challenge provider if none is specified
	if(!credentialProvider) {
		credentialProvider = new jazz.app.internal.LoginDialog();
	}
	
	function cancelAllAndReset() {
		requestManager.cancelPendingRequests();
		challengePending = false;
	}
	
	// If the credential provider does not allow retrying on failure, then
	// cancel all pending requests and reset the challengePending state
	if(state === STATE_PREVIOUS_ATTEMPT_FAILED && !credentialProvider.retryOnFailure) {
		cancelAllAndReset();
		return true;
	}
	
	if(state === STATE_NEW_CHALLENGE) {
		// If this is a challenge to a normal request, pend it to resend after
		// successful log in.
		//
		// NOTE: The first request that got challenged may result in an extra
		// GET on that URL - see Bug 70579 for details.
		setUnauthorized(origRequest.clientDeferred.ioArgs);
		requestManager.pendRequest(origRequest);
		if(challengePending) {
			// If a challenge is already pending, then return (cancelling
			// client callbacks) so that the user doesn't get prompted multiple
			// times.
			return true;
		}
	}
	
	// If we're still running, then we need to challenge the credential provider
	// for credentials.
	
	// Prevent subsequent challenges on normal requests
	challengePending = true;
	
	// Ask for credentials
	credentialProvider.getCredentials({
		onSubmit: function(userId, password) {
			var authRequest = requestManager.createRequest("POST", getJEEFormAuthArgs(userId, password, origRequest.args.url), true);
			var dfd = requestManager.forceRequest(authRequest);
			dfd.addBoth(function(result) {
				if(jazz.app.internal.loginUtil.getUserIdCookie()) {
					jazz.app.internal.loginUtil.setUserIdCookie(userId);
				}
				
				// if this gets called, authentication succeeded (though something else may ahve gone wrong).
				challengePending = false;
				// send all pending requests
				requestManager.sendPendingRequests();
				
				dojo.publish(jazz.app.events.LOGIN, [null]);
			});
		},
		onCancel: function() {
			cancelAllAndReset();
		},
		state: state === STATE_PREVIOUS_ATTEMPT_FAILED ? "login-failed" : null
	});
	
	// return true to cancel the response callbacks
	return true;
}

jazz.app.internal.AuthenticationInterceptor.setCredentialProvider = function(newCredentialProvider) {
	credentialProvider = newCredentialProvider;
};

/**
 * Reports on what information about authentication interactions the response
 * contained, if any. This method will return one of four values:
 * 
 * - STATE_NORMAL_RESPONSE - The response did not contain any information about
 *   authentication interactions.
 * - STATE_NEW_CHALLENGE - The response indicates that the client is being
 *   challenged for credentials for the first time (i.e. not in response to a
 *   failed login attempt).
 * - STATE_PREVIOUS_ATTEMPT_FAILED - The response indicates that the client is
 *   being challenged for credentials in response to previous login attempt that
 *   failed because of invalid credentials.
 * - STATE_LOGGED_OUT - The response indicates that it resulted from a call to
 *   the "log out" service.
 * 
 * @param {Object} response the response object possibly containing information
 *  abuut a log in interaction
 */
function getAuthState(response) {
	
	var authMsg = response.getHeader("X-com-ibm-team-repository-web-auth-msg");
	
	if(!authMsg) {
		// optimize for common case
		return STATE_NORMAL_RESPONSE;
	}
	
	switch(authMsg) {
		case "authrequired":
			return STATE_NEW_CHALLENGE;
		case "authfailed":
			return STATE_PREVIOUS_ATTEMPT_FAILED;
		case "loggedout":
			return STATE_LOGGED_OUT;
		default:
			console.warn("AuthenticationInterceptor: Saw unexpected auth message: " + authMsg);
			return STATE_NORMAL_RESPONSE;
	}
	
}

/**
 * Returns a dojo.__xhrArgs object that will result in a login attempt against
 * JEE Form Based authentication.
 * 
 * @param {Object} userId the user-provided user ID
 * @param {Object} password the user-provided password
 * @param {Object} requestUrl the original URL that was accessed 
 */
function getJEEFormAuthArgs(userId, password, requestUrl) {
	
	var urlPrefix = requestUrl.substring(0, requestUrl.lastIndexOf("/") + 1);
	
	return {
		contentType: "application/x-www-form-urlencoded; charset=utf-8",
		url: urlPrefix + "j_security_check",
		content: {
            j_username: userId,
            j_password: password
        }
	};
}

/**
 * Sets the XHR object's response to indicate that the user was not authorized.
 * This only should be seen by a caller in the case where a log in attempt was
 * cancelled after sending a remote request. 
 * 
 * @param {Object} ioArgs the Dojo ioArgs object
 * @return {null}
 */
function setUnauthorized(ioArgs) {
	var newXhr = {};
	dojo.forEach(["getAllResponseHeaders", "getResponseHeader", "responseText", "responseXML"], function(prop) {
		if(dojo.isFunction(ioArgs.xhr[prop])) {
			newXhr[prop] = dojo.hitch(ioArgs.xhr, ioArgs.xhr[prop]);
		} else {
			newXhr[prop] = ioArgs.xhr[prop];
		}
	});
	dojo.mixin(newXhr, {
		status: 401,
		statusText: "Unauthorized"
	});
	ioArgs.xhr = newXhr;
}

})();